﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Algorithms.DataAnalysis {
  [StorableClass]
  [Item("WeightedEuclideanDistance", "A weighted norm function that uses Euclidean distance √(Σ(w[i]*(p1[i]-p2[i])²)/Σw[i])")]
  public class WeightedEuclideanDistance : DistanceBase<IEnumerable<double>> {
    public const string WeightsParameterName = "Weights";
    public IValueParameter<DoubleArray> WeigthsParameter {
      get { return Parameters[WeightsParameterName] as IValueParameter<DoubleArray>; }
    }

    public DoubleArray Weights {
      get { return WeigthsParameter.Value; }
      set { WeigthsParameter.Value = value; }
    }

    #region HLConstructors & Cloning
    [StorableConstructor]
    protected WeightedEuclideanDistance(bool deserializing) : base(deserializing) { }
    protected WeightedEuclideanDistance(WeightedEuclideanDistance original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new WeightedEuclideanDistance(this, cloner);
    }
    public WeightedEuclideanDistance() {
      Parameters.Add(new OptionalValueParameter<DoubleArray>(WeightsParameterName, "The weights used to modify the euclidean distance. If no weights are specified a Random Forrest Regression / Classification is used to automatically set the weigths. "));
    }
    #endregion

    public static double GetDistance(IReadOnlyList<double> point1, IReadOnlyList<double> point2, DoubleArray impacts) {
      if (point1.Count != point2.Count) throw new ArgumentException("Weighted Euclidean distance not defined on vectors of different length");
      if (impacts == null || impacts.Count() != point1.Count) throw new ArgumentException("Weighted Euclidean distance requires a non-null weight vector of length equal to the number of allowed input double variables the compared points");
      var sum = 0.0;
      var sumW = 0.0;
      for (var i = 0; i < point1.Count; i++) {
        var d = point1[i] - point2[i];
        var w = impacts[i] * impacts[i];
        sum += d * d * w;
        sumW += w;
      }
      return Math.Sqrt(sum / sumW);
    }

    public override double Get(IEnumerable<double> a, IEnumerable<double> b) {
      return GetDistance(a.ToArray(), b.ToArray(), Weights);
    }
  }
}